# Introduction
In the [basics](scl.md#statements) section it has been noted that the order of the statements in the program provide no information for the interpreter and the output will not depend on this order. Rather *the dependencies* between variables and their parameters determine the execution order.

# Side effects

The statements for input/output (I/O) operations, such as `print`, loading from file/URL and exports to file/URL introduce side effects. Without them it is impossible to interact with the program. These effects create situations where a specific ordering of evaluations becomes necessary under certain policies.

# Language interpreters and execution modes

Currently, three different interpreters can be used to execute programs. None of the interpreters performs *eager evaluation*, i.e. evaluates the parameters (such as function calls, expressions etc.) by the interpreter as soon as they are encountered. Rather the evaluation is triggered only when a certain value is actually needed. The order of this evaluation is neither the order of statements in the source code nor the order of interpretation. This is sometimes referred to as *lazy evaluation*.

Another aspect of evaluation is the time of evaluation - during interpreter runs (let us call it *immediate*) and after the interpreter has finished (*deferred* evaluation).

A third aspect is the location of the evaluation: we make difference between *local* and *remote* evaluation. This is when we need additional computing resources that are not available locally (i.e. local HPC cluster with a batch system and a JupyterHub instance connected to it).

The choice of executor determines the *mode* of execution that is selected with `--mode | -m` command-line flags.

## Evaluation policies

In addition to execution mode, we make difference between three different evaluation policies.

### Evaluate none

The policy can only be applied to workflow executor. In this case the iterpreter returns after an instance of the workflow is created or the instance of the workflow is updated.

### Evaluate all

The policy can only be applied to workflow executor. All parameters are evaluated no matter whether they should be printed or not, and after all evaluations are completed, the interpreter executes the print statements in the order of their occurrence.

### Evaluate on demand

Only parameters of `print` statements are evaluated and printed in the order of occurrence in the program.

## Implementation of I/O calls

In the instant and deferred executors, all I/O operations are implemented as blocking calls. This means that the interpreter does not return the control until all parameters of all print statements are evaluated and printed in the order of print statements.

In the workflow executor, the interpreter calls for all I/O statements different from `print` statements are non-blocking, i.e. the interpreter immediately returns after iterpreting them. In contrast, the `print` statements are all implemented with blocking calls, i.e. the interpreter returns the control only after the `print` output has been written on the screen.

### Print operation blocked by parameter evaluation

With the autorun and on-demand evaluation policies the `print` statements are blocked by the evaluation of their parameters if the interpreter acts as workflow engine itself, i.e. performs the actual execution. In these cases, the `print` statements return only when all their parameters have been evaluated.

### Print operation not blocked by parameter evaluation

With no-evaluation policy or if an external workflow engine is used, the `print` statements return immediately. If at the time of the interpretation the parameters have been evaluated their values are printed on the screen, otherwise `n.c.` (*not computed*) is printed.


| interpreter       | location |  time     | policy    | persistence | print calls |
|-------------------|----------|-----------|-----------|-------------|-----------|
| instant executor  | local    | immediate | on demand | no          | blocked  |
| deferred executor | local    | deferred  | on demand | no          | blocked  |
| workflow executor | remote   | deferred  | none, all, on demand | all but not `print` and `vary` | blocked / not blocked |

## Persistence

The instant and deferred executors provide no persistence of the model. Therefore, the model and parts of it cannot be reused and the model cannot be extended. Also no data provenance information is saved.

The workflow executor interprets the model and stores the model instance as a workflow on a database (MongoDB). Provenance metadata is captured and saved and deferred remote or distributed execution is enabled. All statements are persisted except for `print` and `vary` statements. See chapter [Bulk processing](bulk.md) for more about `vary` statements.
