# Supporting tools

To enable the usage of the SCL and AMML, a minimum set of supporting tools is necessary. These tools include a parser, a correctness checker, an interpreter, and a syntax highlighter. All these tools (or subsets of them) can be used via the command-line interface or the Jupyter/iPython kernel.

## Using the command-line interface

The command-line interface is provided by the scripts located in the `scripts` folder.

### Script mode

The command `texts script` runs a DSL *program* (also called a *model*) in one of three different evaluation modes. The basics on how to write a program in the Scientific Computing Language can be found in [this section](scl.md) and Atomic and Molecular Modeling Language in [this section](amml.md).

In all evaluation modes, the model is first parsed using a grammar. The grammar can be optionally specified by the flag `-g GRAMMAR_PATH, --grammar_path`. The default grammar used for parsing is in the file `src/virtmat/language/grammar/virtmat.tx`. Syntax errors are issued in this phase. The interpreter then starts applying static constraints such as type checks. Static errors (such as type errors) are issued at this stage. Then the evaluation of the model is performed. The evaluation (or execution) mode can be specified using the `-m | --mode` option. These modes are *instant*, *deferred*, and *workflow*. The instant evaluation mode is the default. More details are available in the following subsections. The *model source file* (with extension `.vm` ) must always be specified using the flag `-f MODEL_FILE, --model_file MODEL_FILE` even if it is empty.

#### Instant and deferred modes

In the *instant* mode the evaluation is performed *immediately* and *locally*. In *deferred* mode the evaluation is deferred but still performed *locally*. In *instant* and *deferred* modes the model does not persist after the interpreter has finished, i.e. the `texts script` has returned. Only parameters of `print` statements in the program will be evaluated. For debugging purposes, the flag `-p, --print-all-vars` can be specified with which all variables will be evaluated and their names, values and types will be printed.

#### Workflow mode

In the *workflow* mode the intepreter creates a workflow from the model and stores the workflow in a database. Therefore, the model is *persistent* after the interpreter returns, i.e. the *model instance* continues existing on the database. A model instance can be retrieved via a *universally unique identifier* (UUID) that is generated at the time when the model instance is created. A model instance can only be evaluated and extended but not otherwise modified.

If no additional flags are specified the model is not evaluated after interpretation except for parameters in `print` statements that contain no references to variables. Otherwise `n.c.` will be printed instead of the not yet computed values. Optionally, a model can be evaluated locally after the intepreter finishes by specifying the ` -a | --autorun` flag. In this case all variables will be evaluated but only those in print statements will be printed. Additionally, the `-d, --on-demand` flag restricts the evaluation to only those variables that are referenced in `print` statements.

In order to work in workflow mode a launchpad (a MongoDB database) must be configured. Using the flag `-l LAUNCHPAD_FILE, --launchpad_file LAUNCHPAD_FILE` the launchpad data (such as hostname, database name and credentials) can be optionally specified. Otherwise the script looks up the current working directory and then the folder `$HOME/.fireworks` for a launchpad configuration. More information about configuring a launchpad can be found [here](https://gitlab.kit.edu/kit/virtmat-tools/vre-middleware/-/blob/master/docs/launchpad.md).

By using the `-h, --help` flag a summarized help can be obtained:

```
texts script --help
usage: texts script [-h] -f MODEL_FILE [-g GRAMMAR_PATH] [-m {instant,deferred,workflow}] [-p] [-t] [-l LAUNCHPAD_FILE] [-u UUID] [-r] [-d] [-a] [-q QADAPTER_FILE] [--enable_logging] [--logging_level {NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL}]

Run a script

optional arguments:
  -h, --help            show this help message and exit
  -f MODEL_FILE, --model_file MODEL_FILE
                        path to model file
  -g GRAMMAR_PATH, --grammar_path GRAMMAR_PATH
                        path to grammar root file
  -m {instant,deferred,workflow}, --mode {instant,deferred,workflow}
                        execution mode
  -p, --print-all-vars  prints all variables
  -t, --textx-debug     enable textx debug mode
  -l LAUNCHPAD_FILE, --launchpad_file LAUNCHPAD_FILE
                        path to launchpad file (workflow mode only)
  -u UUID, --uuid UUID  UUID of a model to extend/run (workflow mode only)
  -r, --autorun         run the model (workflow mode only)
  -d, --on-demand       run on demand (workflow mode only)
  -a, --async-run       run in background (workflow mode only)
  -q QADAPTER_FILE, --qadapter_file QADAPTER_FILE
                        path to default qadapter file (workflow mode only)
  --enable_logging      enable logging messages
  --logging_level {NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL}
                        logging level
```

A persistent model can be retrieved in workflow mode by using the flag `-u UUID, --uuid UUID`. A model can be optionally evaluated in workflow mode by adding the `-r, --autorun` and `-d, --on-demand` flags or/and extended by specifying a model extension with the flag `-f MODEL_FILE, --model_file MODEL_FILE`. The evaluation with the `-r, --autorun` without the `-a, --async-run` flag is effective only for workflow nodes of [_interactive_ category](resources.md#policies-for-interactive-and-batch-execution).

The flag `-a, --async-run` activates background evaluation using a workflow engine with a launcher thread. The evaluation includes workflow nodes of both [_interactive_ and _batch_ categories](resources.md#policies-for-interactive-and-batch-execution). For this feature the [vre-middleware package](https://gitlab.kit.edu/kit/virtmat-tools/vre-middleware) must be installed. The `WFEngine` object used in the run is dumped into a file in the same folder as the launchpad file. This object can be used independently with the Python API or the GUI of the [vre-middleware package](https://gitlab.kit.edu/kit/virtmat-tools/vre-middleware). The flag `-q, --qadapter_file` is explained in detail [in this section](#evaluation-of-nodes-of-batch-category-custom-qadapter-files).


### Interactive session

The interactive session, started with `texts session`, is a tool managing a session for processing a model in *workflow mode*. Some of the startup flags are the same as for `texts script`.

The flag `-a, --async-run` activates background evaluation using a workflow engine with a launcher thread. For this feature the [vre-middleware package](https://gitlab.kit.edu/kit/virtmat-tools/vre-middleware) must be installed. The `WFEngine` object used in the run is dumped into a file in the same folder as the launchpad file. This object can be used independently with the Python API or the GUI of the [vre-middleware package](https://gitlab.kit.edu/kit/virtmat-tools/vre-middleware). The flag `-q, --qadapter_file` is explained in detail [in this section](#evaluation-of-nodes-of-batch-category-custom-qadapter-files).

```
texts session --help
usage: texts session [-h] [-g GRAMMAR_PATH] [-l LAUNCHPAD_FILE] [-u UUID] [-r] [-d] [-a] [-q QADAPTER_FILE] [--enable_logging] [--logging_level {NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL}]

Start an interactive session

optional arguments:
  -h, --help            show this help message and exit
  -g GRAMMAR_PATH, --grammar_path GRAMMAR_PATH
                        path to grammar root file
  -l LAUNCHPAD_FILE, --launchpad_file LAUNCHPAD_FILE
                        path to launchpad file
  -u UUID, --uuid UUID  UUID of a model
  -r, --autorun         run the model
  -d, --on-demand       run on demand
  -a, --async-run       run in background
  -q QADAPTER_FILE, --qadapter_file QADAPTER_FILE
                        path to default qadapter file
  --enable_logging      enable logging messages
  --logging_level {NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL}
                        logging level
  -w, --ignore_warnings
                        do not display warnings
```

#### Session behavior

In the session, inputs are interactively typed and the outputs are shown immediately. An empty model instance is created at startup and with every input the model instance is extended with the model source extension typed as input. The editing capabilities of the input fields are very limited.

#### Specific features

In the interactive session, if a valid parameter is typed it is interpreted as a `print` statement. For example, a typed input `a` is changed to `print(a)` before parsing. Multiline comments are not supported in the interactive session.

A successfully parsed and interpreted input cannot be modified. It becomes permanently part of the model source.

The session manager accepts additional inputs that are no model sources in SCL or AMML. These are sometimes called magic commands (like in iPython and Jupyter) and begin with the symbol `%`.

magic commands                     | action
-----------------------------------|--------
`%exit`, `%bye`, `%close`, `%quit` | close (quit) the session
`%start`                           | start the background launcher thread
`%stop`                            | stop the background launcher thread
`%uuid`                            | prints the UUID of the active model
`%uuid <UUID>`                     | switch to model with UUID
`%new`                             | create a new session and a new model
`%hist`, `%history`                | prints the current model source

The `%history` command prints all statements (without `print` statements that are not persisted), one per line, with timestamp of most recent evaluation update and current evaluation state. The evaluation state is currently passed through from the FireWorks workflow management system and is described in detail [here](https://materialsproject.github.io/fireworks/reference.html#fireworks-states).


## Using the Jupyter kernel

The Jupyter kernel is still work in progress. Currently, some help can be obtained in [this README document](../virtmat_kernel/README.rst).


## Interpreting the error messages

Either the parser or the interpreter may detect *static errors* in the program. Then they issue specific error messages in the following format:

```
Error type: /absolute/path/to/model_source_file.vm:line:column --> context <--
Detailed error description
```

The `--> context <--` displays the section of the model source where the error has occurred. If the Jupyter kernel or `texts session` is used then `None` is printed instead of an absolute path to file with the model source.

**NOTE:** If an error message is not in this particular format, especially if it is a Python exception with a stack trace beginning with the word "Traceback", then it is most likely due to a bug in the interpreter and not due to an error in the model source. In such cases please [submit an issue](https://gitlab.kit.edu/kit/virtmat-tools/vre-language/-/issues/new) with the tag *Bug* and provide a minimal input to reproduce the error.


### Simple examples of errors

Input:
```
b = 1 a = 2
```
Error:
```
Syntax error: None:1:7 --> b = 1 *a = 2  <--
Expected '[' or '^|;|\,|\)|$' or '**' or '*' or '/' or '+' or '-' or '<=' or '>=' or '>' or '<' or '!=' or '==' or 'if' or 'for' or 'on' or '^|;' or EOF
```

All syntax errors are issued by the parser.

Input:
```
b = a
```
Error:
```
Unknown object: None:1:5
Unknown object "a" of class "GeneralVariable"
```

Unknown object error is a semantic error issued by the parser when it cannot resolve a reference.

Input:
```
s = 1
s = 'hello'
```
Error:
```
Initialization error: None:2:1 --> s = 'hello' <--
Repeated initialization of "s"
```

Repeated initialization error is issued by the interpreter when it applies semantic constraints before the beginning of evaluation.

### Runtime errors

Some errors occur only during the evaluation. These are called *runtime errors*. Runtime errors have the same format as the static errors. Runtime errors only occur if the evaluation is requested, e.g. by `print` statements. For example, this input:

```
length = 1 [m]
b = length + 1
```

will produce no error message. Adding `print(b)` to the model and running the model in instant or deferred evaluation mode will yield this error:

```
Dimensionality error: None:2:5 --> length + 1 <--
Cannot convert from 'meter' ([length]) to 'dimensionless' (dimensionless)
```

In workflow evaluation mode this error occurs only if the evaluation is enabled and will be issued only if there is `print(b)` statement.


## Evaluation of nodes of batch category: custom _qadapter_ files

Once a model has been created in workflow mode, thus added to the database, it can be evaluated using either `texts script` or `texts session` with the `--autorun, -r` flag. In this mode, the workflow nodes of _interactive_ category are evaluated synchronously while the nodes of _batch_ category are not. More information about the workflow node categories can be found [here](resources.md#policies-for-interactive-and-batch-execution). To enable asynchronous background evaluation of batch-category nodes one has to add the `--async-run, -a` flag.
The background evaluation of the model involves the activation of the optional `--qadapter, -q` flag, which takes the name of a _qadapter_ file (e.g. qadapter.yaml) as argument.
Such file provides the details about the computing resources required to run a node.
This is particularly important for nodes of _batch_ category, as this _qadapter_ specifies queue managing system, queue name, ammount of memory, number of processors, nodes, run time, etc.
The _qadapter_ can optionally specify any pre- and post-run requirements (such as loading software modules, or setting environment variables), as well as the execution of the main task.

A simple _qadapter_ can be automatically generated from the model's [resource annotations](./resources.md#resource-annotation).
This means **only resource specifications** (computing time, number of nodes, number of cores, number of GPUs, and ammount of memory) are written to the _qadapter_.

**Any additional requirements have to be defined manually by the user.**
This can be achieved with relative ease by including the special requirements alongside the queue, and computing resource specifications in a custom _qadapter file_.
The pre- and post-run requirements should be defined inside the `pre_rocket` and `post_rocket` keys of the custom _qadapter file_.
The key `rocket_launch` defines the command of the main process and allows some fine tuning: for example, if a non-default launchpad is used then the user must provide the full path to the launchpad file via the `-l` flag.

When the evaluation run is performed, the path and name of the _qadapter file_ should be specified as arguments of the `-q` option of the CLI-based executors.


#### Example: custom _qadapter_ that loads software modules

As an example, consider a hypothetical model that requires the evaluation of some properties employing a domain-specific code called `vasp`.
Execution of this code needs computing resources assigned through a batch system.
Let us asume that the model has the following resource requirements:
- run on 10 cores;
- allow the `vasp` job to run for 20 minutes.

These can be specified by the following resource annotation in the property object of the model:
```
on 10 cores for 20.0 [minutes]
```

Additionally to the resource assignment, the following pre-run requirements should be satisfied to complete the evaluation:
- purge all software modules in the computing node;
- load, as module, version `6.3.2_version` of a program called `vasp` from a library called `chem`;
- activate the `vre-language` python virtual environment;
- export the path to a custom python library located in `~/my_project/`;
- run the corresponding batch job employing the resources from an account called `example_account`.


If the evaluation was done using, for example, `$> texts session -u IUUD_of_my_model -a -r`, the session manager would generate a **basic  qadapter file** similar to:
```
_fw_name: CommonAdapter
_fw_q_type: SLURM
_fw_timeout: 5
cpus_per_task: 1
nodes: 1
ntasks: 1
ntasks_per_node: 1
queue: dev_cpuonly
rocket_launch: rlaunch singleshot
walltime: 20
```
which includes no reference to the pre-run requisites nor the account providing the computing time!

In contrast, a custom _qadapter_ allows users to have much more control over the resources and specifications needed to run the workflow nodes.
For the current sample case, it would look something like:
```
_fw_name: CommonAdapter
_fw_q_type: SLURM
_fw_timeout: 5
cpus_per_task: 1
nodes: 1
ntasks: 1
ntasks_per_node: 1
queue: dev_cpuonly
account: example_account
pre_rocket: module purge
            module load chem/vasp/6.3.2_version
            . /home/python_VirtEnvs/vre-language/bin/activate
            export PYTHONPATH=/home/my_project/lib:$PYTHONPATH
rocket_launch: rlaunch singleshot
walltime: 20
```
By setting the `pre_rocket` key, all pre-requisites are satified before running the evaluation task.
Also, the number of cores (10) will be assigned directly in the batch run script of the VASP job.


The new _qadapter.yaml_ file should be provided as argument of the `-q` flag of the `texts` tool.
As an example, the model can be fully evaluated using the CLI interactive session by entering:
```
$> texts session -u IUUD_of_my_model -a -r -q path/to/my/custom/qadapter.yaml
```
