"""interactive session"""
import os
import warnings
from argparse import ArgumentParser
from fireworks.fw_config import LAUNCHPAD_LOC
from fireworks.scripts.lpad_run import get_lp
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter
from virtmat.language.interpreter.session_manager import SessionManager
from virtmat.language.utilities.errors import error_handler
from virtmat.language.utilities.textx import GRAMMAR_LOC
from virtmat.language.utilities import logging


def parse_clargs():
    """parse the command line arguments for the script"""
    m_description = 'Execute a program remotely in deferred mode'
    parser = ArgumentParser(description=m_description)
    add_arguments(parser)
    return parser.parse_args()


def add_arguments(parser):
    """add arguments to a parser object"""
    parser.add_argument('-g', '--grammar_path', required=False,
                        help='path to grammar root file', default=GRAMMAR_LOC)
    parser.add_argument('-l', '--launchpad_file',
                        help='path to launchpad file', default=LAUNCHPAD_LOC)
    parser.add_argument('-u', '--uuid', default=None, required=False,
                        help='UUID of a model')
    parser.add_argument('-r', '--autorun', default=False, required=False,
                        action='store_true', help='run the model')
    parser.add_argument('-d', '--on-demand', default=False, required=False,
                        action='store_true', help='run on demand')
    parser.add_argument('-a', '--async-run', default=False, required=False,
                        action='store_true', help='run in background')
    parser.add_argument('-q', '--qadapter_file', default=None, required=False,
                        help='path to default qadapter file')
    parser.add_argument('--enable_logging', default=False, required=False,
                        action='store_true', help='enable logging messages')
    parser.add_argument('--logging_level', default='CRITICAL', required=False,
                        choices=logging.LOGGING_LEVELS, help='logging level')
    parser.add_argument('-w', '--ignore_warnings', default=False, required=False,
                        action='store_true', help='do not display warnings')


@error_handler
def main(clargs):
    """main function"""
    logging.LOGGING_LEVEL = logging.get_logging_level(clargs.logging_level)
    if clargs.ignore_warnings:
        warnings.filterwarnings('ignore')
    if not clargs.enable_logging:
        logging.disable_logging()
    config_dir = os.path.dirname(os.path.abspath(clargs.launchpad_file))
    qadapter = CommonAdapter.from_file(clargs.qadapter_file) if clargs.qadapter_file else None
    mgr = SessionManager(get_lp(clargs), uuid=clargs.uuid, autorun=clargs.autorun,
                         grammar_path=clargs.grammar_path, on_demand=clargs.on_demand,
                         async_run=clargs.async_run, config_dir=config_dir,
                         qadapter=qadapter)
    mgr.main_loop()


if __name__ == '__main__':
    main(parse_clargs())
