"""apply various constraints for literal values"""
from virtmat.language.utilities.errors import textxerror_wrap, StaticValueError
from virtmat.language.utilities.resources import res_config


@textxerror_wrap
def check_resources_values(obj):
    """object processor to check Resources objects for general validity"""
    if obj.ncores is not None and obj.ncores < 1:
        raise StaticValueError('number of cores must be a positive integer number')
    if obj.walltime is not None and obj.walltime.value <= 0:
        raise StaticValueError('walltime must be a positive number')
    if obj.memory is not None and obj.memory.value <= 0:
        raise StaticValueError('memory must be a positive number')


@textxerror_wrap
def check_resources_values_limits(obj):
    """check whether requested Resources are within the available resources"""
    if obj.ncores is not None and obj.ncores > res_config['max_ncores']:
        raise StaticValueError('number of cores exceeds maximum available resources')
    if obj.walltime is not None and obj.walltime.value > res_config['max_walltime']:
        raise StaticValueError('walltime exceeeds the maximum allowed walltime')
    if obj.memory is not None and obj.memory.value > res_config['max_memory']:
        raise StaticValueError('memory exceeds the maximum allowed memory')


@textxerror_wrap
def check_number_of_chunks(obj):
    """check number of chunks in parallelizable objects"""
    if obj.nchunks is not None and obj.nchunks < 1:
        raise StaticValueError('number of chunks must be a positive integer number')
