"""set property and further class attributes to metamodel classes"""
from textx import textx_isinstance
from virtmat.language.constraints.typechecks import add_type_properties
from virtmat.language.interpreter.instant_executor import add_value_properties
from virtmat.language.interpreter.deferred_executor import add_deferred_value_properties
from virtmat.language.interpreter.workflow_executor import add_workflow_properties
from virtmat.language.interpreter.deferred_executor import add_func_properties
from virtmat.language.utilities.logging import get_logger
from .function import add_function_call_properties
from .clone import add_deepcopy
from .units import add_units_properties


def add_metamodel_classes_attributes(metamodel):
    """add various attributes to metamodel classes"""

    def get_column(self, name):
        """get a column with given name from a table, return series"""
        ret = None
        for col in self.columns:
            if textx_isinstance(col, metamodel['Series']):
                if col.name == name:
                    ret = col
                    break
        return ret

    metamodel['Table'].get_column = get_column


def add_properties(metamodel, deferred_mode=False, workflow_mode=False):
    """set attributes and properties in the metamodel"""
    add_model_param_defs(metamodel)
    add_deepcopy(metamodel)
    add_metamodel_classes_attributes(metamodel)
    add_type_properties(metamodel)
    if deferred_mode or workflow_mode:
        get_logger(__name__).info('running in deferred_mode')
        add_func_properties(metamodel)
        if workflow_mode:
            get_logger(__name__).info('running in workflow mode')
            add_workflow_properties(metamodel)
        else:
            add_deferred_value_properties(metamodel)
    else:
        add_value_properties(metamodel)
    add_units_properties(metamodel)
    add_function_call_properties(metamodel)


def add_model_param_defs(metamodel):
    """add model parameter definitions to access with model._tx_model_params"""
    metamodel.model_param_defs.add('model_instance', 'Model instance')
    metamodel.model_param_defs.add('source_code', 'Source code of the model')
    metamodel.model_param_defs.add('grammar_str', 'Grammar for parsing the model')
    metamodel.model_param_defs.add('deferred_mode', 'Deferred evaluation mode')
    metamodel.model_param_defs.add('autorun', 'Run the model')
    metamodel.model_param_defs.add('on_demand', 'Run the model nodes on demand')
