"""shared fixtures requested by tests in several different modules"""
import pytest
from fireworks import LaunchPad
from fireworks.fw_config import LAUNCHPAD_LOC
from textx import metamodel_from_file
from virtmat.language.metamodel.properties import add_properties
from virtmat.language.metamodel.processors import add_processors
from virtmat.language.utilities.textx import GRAMMAR_LOC


@pytest.fixture(name='test_config',
                params=[(False, False), (True, False), (True, True)],
                ids=['INSTANT', 'DEFERRED', 'WORKFLOW'])
def fixture_config(request):
    """
    params is defined as (DEFERRED_MODE, WORKFLOW_MODE)
    with the 3 parameters the other fixtures and every test are parametrised
    differently and executed three times
    """
    return request.param


@pytest.fixture(name='model_kwargs')
def fixture_kwargs(test_config):
    """setting up kwargs for model instance"""
    model_kwargs = {'deferred_mode': test_config[0]}
    if test_config[1]:
        lpad = LaunchPad.from_file(LAUNCHPAD_LOC) if LAUNCHPAD_LOC else LaunchPad()
        model_kwargs['model_instance'] = {'lpad': lpad, 'uuid': None}
        model_kwargs['autorun'] = True
    return model_kwargs


@pytest.fixture(name='raw_meta_model')
def fixture_raw_metamodel():
    """create an initial metamodel from a grammar file"""
    return metamodel_from_file(GRAMMAR_LOC, auto_init_attributes=False)


@pytest.fixture(name='meta_model')
def fixture_metamodel(test_config, raw_meta_model):
    """
    add custom properties and processors to an initial metamodel and
    return a modified metamodel
    """
    add_properties(raw_meta_model, deferred_mode=test_config[0],
                   workflow_mode=test_config[1])
    add_processors(raw_meta_model, wflow_processors=test_config[1])
    return raw_meta_model


@pytest.fixture(name='model_kwargs_wf')
def fixture_kwargs_wf():
    """setting up kwargs for model instance for the workflow execution mode"""
    model_kwargs = {'deferred_mode': True}
    lpad = LaunchPad.from_file(LAUNCHPAD_LOC) if LAUNCHPAD_LOC else LaunchPad()
    model_kwargs['model_instance'] = {'lpad': lpad, 'uuid': None}
    model_kwargs['autorun'] = True
    return model_kwargs


@pytest.fixture(name='meta_model_wf', params=[True], ids=['WORKFLOW'])
def fixture_metamodel_wf(raw_meta_model):
    """modified metamodel for the workflow execution mode"""
    add_properties(raw_meta_model, deferred_mode=True, workflow_mode=True)
    add_processors(raw_meta_model, wflow_processors=True)
    return raw_meta_model
